# Copyright 2015-2022 - RoboDK Inc. - https://robodk.com/
#from .robodk_api import *
#
#from .robodk_api.robodk import *
#from .robodk_api.robolink import *

from enum import IntEnum

#from robolink import *
#from robodk import *

from . import robodk
from . import robolink

import time

from .fission.utils.message_box import message_box
import tempfile
import adsk.core, adsk.fusion



global DEBUG_PLUGIN
DEBUG_PLUGIN = True
   
def printLog(msg):
    global DEBUG_PLUGIN
    if DEBUG_PLUGIN:
        #import sys
        print(msg)
        #sys.stdout.flush()
        open_mode = 'a'
        if type(DEBUG_PLUGIN) is not str:            
            DEBUG_PLUGIN = tempfile.gettempdir() + '/RoboDK.Fusion360.debug.txt'
            open_mode = 'w'
        with open(DEBUG_PLUGIN, open_mode) as f:
            f.write(msg)
            f.write('\n')
    
    
class CurveDisplayEnum(IntEnum):
    Default = 0
    Yes = 1
    No = 2
    
    
class UpdateLevelEnum(IntEnum):
    NoUpdate = 0          # Just send the geometry to RoboDK
    Update = 1            # Send the geometry and update (default)
    UpdateAndSimulate = 2 # Send the geometry, update and start simulation
    UpdateAndGenerate = 3 # Send the geometry, update and generate program
    UpdateAndSend = 4     # Send the geometry, update and send program to robot
    
    
STATUS_ERROR = -1 # An error message is displayed
STATUS_STOP = 0   # We need to stop loading stuff to RoboDK
STATUS_OK = 1     # All good

def strip_unsafe_chars_for_filename(filename):
    keepcharacters = (' ','.','_')
    return "".join(c for c in filename if c.isalnum() or c in keepcharacters).rstrip()

# Same utilities available in RoboDK for vectors but using a point as an XYZ struct (not an array)
def adsk_norm(p):
    """Returns the norm of a 3D vector"""
    return sqrt(p.x*p.x + p.y*p.y + p.z*p.z)
   
def adsk_distance(a,b):
    """Calculates the distance between two points"""
    return a.distanceTo(b)

def Points_2_Mat(settings, in_point_list, in_face_list, adsk_product=None):
    """Calculate the normal vector for each point with respect to the selected surfaces 
    and convert a list of points with their normal vector to a list of Mat
    
    :param settings: User defined settings
    :param list in_point_list: List of points [X,Y,Z]
    :param list in_point_list: List of faces custom structure
    :param product: do we need it? (TODO)
    """
    
    out_list_mat = []
    npoints = len(in_point_list)
    nfaces = len(in_face_list) if in_face_list else 0
    calc_normals = nfaces > 0

    curve_points = robodk.Mat(6 if calc_normals else 3, npoints)
    xyz_start = 0
    xyz_end = npoints
    xyz_step = 1
    if settings.CurveZigZag:
        xyz_start = npoints - 1
        xyz_end = -1
        xyz_step = -1
    
    xyz = xyz_start
    xyz_count = 0
    while xyz != xyz_end:
        pti = in_point_list[xyz_count]
        # Drawing units are in CM. RoboDK needs units in MM
        curve_points[0, xyz] = 10.0 * pti.x
        curve_points[1, xyz] = 10.0 * pti.y
        curve_points[2, xyz] = 10.0 * pti.z
        if calc_normals:
            normal = Calculate_Normal(settings, pti, in_face_list, adsk_product)
            curve_points[3, xyz] = normal[0]
            curve_points[4, xyz] = normal[1]
            curve_points[5, xyz] = normal[2]

        xyz = xyz + xyz_step
        xyz_count = xyz_count + 1

    out_list_mat.append(curve_points)
    return out_list_mat
    


def Curves_2_Mat(settings, in_curve_list, in_face_list, adsk_product=None):
    """Convert a list of curves to a list of Mat. 
    The curve is split (meshed, or converted to a "polyline" according to desired tolerances).
    :param list in_curve_list: list of curves, each curve is a list of points
    """
    
    ncurves = len(in_curve_list)
    nfaces = len(in_face_list) if in_face_list else 0

    # Define a list of Mat (matrix) that will be fed to RoboDK
    # Each Mat is 3 x np or 6 x np if we selected surfaces
    out_list_mat = []
    
    # Ignore extracting surface normals if no faces are selected
    if nfaces == 0:
        for j in range(0, ncurves):
            # Retrieve each curve
            curve_adsk = in_curve_list[j]
            np = len(curve_adsk)
            
            # Create a new Mat object and iterate through each point in the curve
            curve_mat = robodk.Mat(3, np)
            for i in range(0, np):
                pti = curve_adsk[i]
                curve_mat[0, i] = 10.0 * pti.x
                curve_mat[1, i] = 10.0 * pti.y
                curve_mat[2, i] = 10.0 * pti.z

            out_list_mat.append(curve_mat)

        return out_list_mat

    for j in range(0, ncurves):
        # Retrieve each curve
        curve_adsk = in_curve_list[j]
        np = len(curve_adsk)
        
        # Create a new Mat object and iterate through each point in the curve
        curve_mat = robodk.Mat(6,np)
        for i in range(0,np):
            pti = curve_adsk[i]
            curve_mat[0, i] = 10.0 * pti.x
            curve_mat[1, i] = 10.0 * pti.y
            curve_mat[2, i] = 10.0 * pti.z
            # Calculate surface normal on this point pti
            normal = Calculate_Normal(settings, pti, in_face_list, adsk_product)
            curve_mat[3, i] = normal[0]
            curve_mat[4, i] = normal[1]
            curve_mat[5, i] = normal[2]
            
        out_list_mat.append(curve_mat)
        
    return out_list_mat
    

def Calculate_Normal(settings, point_abs, in_face_list, adsk_product):
    """Calculate the normal of a point point given a list of surfaces. 
    The point and normal should be in absolute part/assembly coordinates"""
    min_dist_combine = 1e9
    min_dist_reset = 1e9
    pti_normal = [0.0,0.0,1.0]
    pt_closest = [0.0,0.0,0.0]

    printLog("POINTABS(:,end+1)=[%.3f, %.3f, %.3f]';" % (10*point_abs.x, 10*point_abs.y, 10*point_abs.z))

    normal_cacl_tolerance = settings.NormalCalcToleranceMM
    # Among all selected surfaces find the surface closest to the tess point
    countface = 0
    ptOnsurf = None
    for faceinfo in in_face_list:    
        # Get the projection of point:
        # TODO: Make sure point_abs is in absolute coordinates and the face evaluator returns absolute coordinates
        #message_box(face.objectType)
        # m = adsk.core.Matrix3D.create()
        # m.getCell(0,0) # same order as Matlab
        # m.setCell(0,3)
        face = faceinfo[0]
        facematrix = faceinfo[1]
        facematrix_inv = faceinfo[2]
                
        pt_rel = point_abs.copy()
        #pt_rel.transformBy(facematrix_inv)
    
        countface += 1
        #https://help.autodesk.com/view/fusion360/ENU/?guid=GUID-31dad9b2-ee1d-4216-8100-09ea44f9967a
        res, param = face.evaluator.getParameterAtPoint(pt_rel)
        if not res:
            message_box("Failed to find point on face near given point.")
            continue

        #https://help.autodesk.com/view/fusion360/ENU/?guid=GUID-e6682eb7-48f8-472d-9872-71da3c5ba8c8        
        #if not face.evaluator.isParameterOnFace(param):
        #    printLog("Point is not on face")
        #    continue
        
        res, ptOnsurf = face.evaluator.getPointAtParameter(param)
        if not res:
            message_box("Failed to find point on face near given point.")
            continue
        
        # Calculate the distance to the surface
        # ptOnSurf is in CM, we need to change to MM
        dist_mm = 10.0 * adsk_distance(ptOnsurf, pt_rel)
        
        #ptOnsurf = ptOnsurf.copy()
        #ptOnsurf.transformBy(facematrix)
        #dist_mm = 10.0 * adsk_distance(ptOnsurf, pt_rel)
        
        printLog("    PTSURF(:,end+1)=[%.3f, %.3f, %.3f]'; %%  Face %i  distance=%.3f mm" % (10*ptOnsurf.x, 10*ptOnsurf.y, 10*ptOnsurf.z, countface, dist_mm))        
        
        if dist_mm < min_dist_combine:
            # get normal of the closest point on the surface 
            res, pti_normal_i = face.evaluator.getNormalAtParameter(param)
            if not res:
                message_box("Failed to compute normal at point.")
                continue
            
            #pti_normal_i.transformBy(facematrix)
            
            # If the projection is much closer than a previous projection, take this normal as the new normal
            if dist_mm < min_dist_reset:
                # Calculate the distance required to replace the normal or combine it
                min_dist_reset = dist_mm - normal_cacl_tolerance
                min_dist_combine = dist_mm + normal_cacl_tolerance
                pti_normal[0] = pti_normal_i.x
                pti_normal[1] = pti_normal_i.y
                pti_normal[2] = pti_normal_i.z
                pt_closest = [ptOnsurf.x, ptOnsurf.y, ptOnsurf.z]
                printLog("    % Closest point!!")
                
            else:
                # If the projection is similar to another point:
                # combine the 2 normals (RoboDK will normalize the vector if it is not a unit vector)
                # This is useful for extracting the normal bisectiong 2 surfaces
                pti_normal[0] += pti_normal_i.x
                pti_normal[1] += pti_normal_i.y
                pti_normal[2] += pti_normal_i.z
                printLog("    %% Combining normal with closest point")

    if settings.InvertNormals:
        pti_normal[0] *= -1.0
        pti_normal[1] *= -1.0
        pti_normal[2] *= -1.0

    #if ptOnsurf:
    printLog("  PTCLOSEST(:,end+1)=[%.3f, %.3f, %.3f]';" % (10*pt_closest[0], 10*pt_closest[1], 10*pt_closest[2]))
    printLog("  NORMAL(:,end+1)=[ %.3f, %.3f, %.3f]'; \n" % (pti_normal[0], pti_normal[1], pti_normal[2]))

    return pti_normal
    

# Start a new instance of the RoboDK API. Returns None if something is not good
def GetRoboDK(settings=None):
    if not robolink.RoboDKInstallFound():
        message = "RoboDK Software not installed.<br><br>Download and install RoboDK from our website<br><a href='https://robodk.com/?RDK&utm_source=RoboDKFusion&utm_medium=Plugin'>https://robodk.com/</a><br><br>Make sure to install RoboDK in the following location:<br>" + getPathRoboDK()
        # TODO: Automatically open RoboDK download page
        #System.Diagnostics.Process.Start("https:#robodk.com/download?RDK&utm_source=RoboDKFusion&utm_medium=Plugin")
        return None, message
            
    RDK = None
    if settings is not None:
        arguments = settings.RDK_ARGS
        port = settings.RDK_COM
        if port is None or port <= 0:
            port = None
        RDK = robolink.Robolink("localhost", port=port, args=arguments)
    else:
        RDK = robolink.Robolink()
        
    if not RDK.Connect():            
        message = "Can't connect to the RoboDK API. Manually start the RoboDK application."
        return None, message

    if settings is not None:
        if settings.ThemeDefaults:
            # Setup Fusion default template (background color, mouse navigation, etc)
            RDK.Command("DisplayCurves", "1")
            RDK.Command("FUSION", "")
        else:
            RDK.Command("ROBODKDEFAULTS", "")
        
    return RDK, ""

#def AdskSavePart(body, file_path, file_format):
#    exporter = adsk.core.Application.get().activeProduct.exportManager
#    component = body.parentComponent if isinstance(body, adsk.fusion.BRepBody) else body
#    opts = {
#        ExportFormatEnum.STEP: lambda: exporter.createSTEPExportOptions(file_path, component),
#        ExportFormatEnum.IGES: lambda: exporter.createIGESExportOptions(file_path, component),
#        ExportFormatEnum.STL: lambda: exporter.createSTLExportOptions(body, file_path)
#    }[file_format]()
#    return exporter.execute(opts)
#    
#def AdskPart2RoboDK(settings, body):
#    """Export the current part/assembly and load it to RoboDK, returns the object as RoboDK.Item. Returns None and a message if failed"""
#    if not body:
#        body = adsk.core.Application.get().activeProduct.rootComponent
#    file_format = settings.ExportFormat
#    if file_format != ExportFormatEnum.Default:
#        file_extension = file_format.name.lower()
#    else:
#        file_format = ExportFormatEnum.STEP
#        file_extension = 'step'
#
#    file_path = tempfile.gettempdir() + strip_unsafe_chars_for_filename(body.name) + "." + file_extension
#    if not AdskSavePart(body, file_path, file_format):
#        message = "Unable to export 3D model to " + file_path
#        return None, message
#
#    item_object, message = Part_2_RoboDK(settings, file_path)
#    return item_object, message
    
def Part_2_RoboDK(settings, file, scale=1.0):
    """Load a file from the temporary path and place it in a reference."""

    # Connect to RoboDK
    RDK, message = GetRoboDK(settings)
    if RDK is None:
        return None, message 

    # Try to get the existing object with matching name
    item_existing = RDK.Item(settings.getNameObject(), robolink.ITEM_TYPE_OBJECT)
    reference = None
    if settings.ObjectOverride and item_existing.Valid():
        reference = item_existing.Parent()
        item_existing.Delete()
    
    # try to get the reference frame
    if reference is None or reference.Type() != robolink.ITEM_TYPE_FRAME:
        reference = RDK_Get_RobotReference(settings, RDK)
        
    # Make sure RoboDK does not extract the curves included in the STEP or IGES file (slower)
    file_lw = file.lower()
    if file_lw.endswith("stp") or file_lw.endswith("step") or file_lw.endswith("igs") or file_lw.endswith("iges"):
        RDK.Command("ImportCurvesAndPointsRead", "0")
        RDK.Command("ImportCurvesAndPointsExtract", "0")

    # Change RoboDK settings to display curves in objects that have geometry
    if settings.ShowObjectCurves.value > 0:
        RDK.Command("DisplayCurves", "1" if settings.ShowObjectCurves.value == 1 else "0")

    # add the file to the reference frame
    item = RDK.AddFile(file, reference)
    if not item.Valid():
        message = "Failed to load file: " + file
        return None, message
    
    if scale != 1.0:
        item.Scale(scale)
        
    RDK.Command("FitAll") # same as selecting Fit All in RoboDK
    message = "File loaded successfully"
    return item, message

def ItemValid(itm):
    """Check if the item is not null and valid"""
    return itm is not None and itm.Valid()

def RDK_Get_RobotReference(settings, rdk):
    """Get or create a new reference frame, relative to a robot base, to place the a part"""
    item_base = None
    default_ref_name = "Part Reference"

    # Try to get an existing reference frame (name match)
    frame_base_name = default_ref_name if settings.NameRef is "" else settings.NameRef

    item_base = rdk.Item(frame_base_name, robolink.ITEM_TYPE_FRAME)
    if item_base.Valid():
        return item_base
    
    # Try to get/add a robot reference frame
    item_robot = rdk.Item(settings.NameRobot, robolink.ITEM_TYPE_ROBOT)

    if item_robot.Valid():
        item_base = item_robot.getLink(robolink.ITEM_TYPE_FRAME)
        if item_base.Valid():
            # The robot has an "active" reference
            return item_base
        
        # Add a reference frame arbitrarily, attached to the robot
        robot_base = item_robot.Parent()

        if robot_base.Type() == robolink.ITEM_TYPE_FRAME:
            if not item_base.Valid() and robot_base.Valid():
                item_base = rdk.AddFrame(frame_base_name, robot_base)
                item_base.setPose(robodk.transl(500, 500, 0))
                return item_base

    # Add a reference frame anyway, at the root of the station, even if we did not find a robot
    item_base = rdk.AddFrame(frame_base_name)
    item_base.setPose(robodk.transl(1000, 0, 0))
    return item_base

def RDK_LoadProject(rdk, rdk_project_file):
    """Check if the desired RDK project has been loaded in RoboDK. If not, load it.
    Returns True if successful"""
    if rdk_project_file is "":
        # Skip if the user didn't specify an RDK file
        return True, ""

    # Get the RDK project file currently loaded in RoboDK
    file_rdkstation = rdk.getParam("FILE_OPENSTATION")
    if rdk_project_file not in file_rdkstation:        
        import os
        if not os.path.isfile(rdk_project_file):
            message = "The RoboDK project file does not exist (" + rdk_project_file + "). Load this project or do not specify a RoboDK project to use the currently open project in RoboDK."
            return False, message

        stn = rdk.AddFile(rdk_project_file)
        if not stn.Valid():
            message = "The RoboDK file (RDK) is not valid. Ignore the RoboDK station parameter or provide the path of a valid file."
            return False, message

    return True, ""

def Mat_2_RoboDK(settings, list_mat, load_as_points = False, item_object = None):
    """Send a list of Mat (curves or points) to RoboDK. The geometry will be added to an existing object if the user specified so (settings)."""
    # connect to RoboDK
    RDK, message = GetRoboDK(settings)
    if RDK is None:
        return None, STATUS_ERROR, message

    # Load the RDK project
    success, message = RDK_LoadProject(RDK, settings.NameRDK)
    if not success:
        return None, STATUS_ERROR, message

    # if the flag first_is added is FALSE it will create a new part
    add_to_existing_object = False
    if item_object is not None:
        add_to_existing_object = True

    # retrieve all the items
    item_object = item_object if item_object is not None else RDK.Item(settings.getNameObject(), robolink.ITEM_TYPE_OBJECT)
    item_robot = RDK.Item(settings.NameRobot, robolink.ITEM_TYPE_ROBOT)
    item_ref = None if len(settings.NameRef) == 0 else RDK.Item(settings.NameRef, robolink.ITEM_TYPE_FRAME)
    item_tool = None if len(settings.NameTool) == 0 else RDK.Item(settings.NameTool, robolink.ITEM_TYPE_TOOL)
    item_milling = None # do not specify a milling project.

    # skip this step if we provided an object externally
    if item_object.Valid():
        item_milling = item_object.getLink(robolink.ITEM_TYPE_MACHINING)
        if item_milling.Valid():
            if not ItemValid(item_ref):
                item_ref = item_milling.getLink(robolink.ITEM_TYPE_FRAME)
            
            if not ItemValid(item_tool):
                item_tool = item_milling.getLink(robolink.ITEM_TYPE_TOOL)
            
        else:
            item_milling = None
        
        # get the object parent if we did not find a valid reference frame
        if not ItemValid(item_ref):
            if ItemValid(item_object):
                item_ref_candidate = item_object.Parent()
                if item_ref_candidate.Type() == robolink.ITEM_TYPE_FRAME:
                    item_ref = item_ref_candidate
        
        if not add_to_existing_object:
            item_object.Delete()
            item_object = None

    # try to get a robot base frame if we do not have a valid reference frame at this point
    if not add_to_existing_object and not ItemValid(item_ref):
        item_ref = RDK_Get_RobotReference(settings, RDK)# item_robot, settings.NameRef.Length > 0 ? settings.NameRef : "SW Reference 1")
    
    # iteratively add all the curves to the object
    for mat in list_mat:
        if load_as_points:
            item_object = RDK.AddPoints(mat, item_object, add_to_existing_object, robolink.PROJECTION_NONE)
        else:
            item_object = RDK.AddCurve(mat, item_object, add_to_existing_object, robolink.PROJECTION_NONE)
            
        if not add_to_existing_object:
            item_object.setName(settings.getNameObject())
            if ItemValid(item_ref):
                item_object.setParent(item_ref)

            add_to_existing_object = True

    # Update the program
    status, message = UpdateProgram(settings, RDK, item_robot, item_ref, item_tool, item_object, "", item_milling)
    return item_object, status, message



def UpdateProgram(settings, rdk, item_robot, item_ref, item_tool, item_object, file_nc, item_mill):
    """Update and generate the program according to settingsions
    progstatus returns how much of the path can be accomplished as a percentage"""
    message = ''
    if settings.UpdateLevel >= UpdateLevelEnum.Update:
        if not ItemValid(item_mill):
            name_mill = settings.getNameObject() + " Settings"
            if settings.NameProgram:
                name_mill = settings.NameProgram + " Settings"

            item_mill = rdk.Item(name_mill, robolink.ITEM_TYPE_MACHINING)
            if not ItemValid(item_robot):
                message = "There are no robots available in the RoboDK station. Load a robot first by selecting File-Open. "
                if settings.NameRobot:
                    message = "The robot " + settings.NameRobot + " doesn't exist in the RoboDK station. Leave the Robot name empty (settings menu) to use the first robot available."
                return STATUS_STOP, message

            if not item_mill.Valid():
                item_mill = rdk.AddMachiningProject(name_mill, item_robot)

        if not item_mill.Valid():
            message = "Unknown problems adding a curve follow project in RoboDK."
            return STATUS_ERROR, message

        program = None
        item_mill.setMachiningParameters(file_nc, item_object)
        if ItemValid(item_ref):
            item_mill.setPoseFrame(item_ref)

        if ItemValid(item_tool):
            item_mill.setPoseTool(item_tool)

        prog_status_out = item_mill.Update(1 if settings.CheckCollisions else 0, 3600)
        prog_status = prog_status_out[3]
        message = rdk.LAST_STATUS_MESSAGE
        if prog_status >= 0.0:
            program = item_mill.getLink(robolink.ITEM_TYPE_PROGRAM)

        #prog_status = prog_status > 0 ? 100 * prog_status : prog_status
        if prog_status <= 0:
            message = "Can't simulate or generate programs because the robot can't reach the start point"
            return STATUS_STOP, message

        if settings.UpdateLevel == UpdateLevelEnum.UpdateAndSimulate:
            if prog_status > 0.0:
                program.RunProgram()

        elif settings.UpdateLevel == UpdateLevelEnum.UpdateAndGenerate:
            if prog_status > 0.0:
                prog_made_ok = program.MakeProgram(settings.NameProgram, robolink.RUNMODE_MAKE_ROBOTPROG)
                message = rdk.LAST_STATUS_MESSAGE # flags any upload errors

        elif settings.UpdateLevel == UpdateLevelEnum.UpdateAndSend:
            if prog_status > 0.0:
                prog_sent_ok = program.MakeProgram(settings.NameProgram, robolink.RUNMODE_MAKE_ROBOTPROG_AND_UPLOAD)
                message = rdk.LAST_STATUS_MESSAGE # flags any upload errors

    else:
        message = "Success! Update the project in RoboDK (Ctrl+U) to check if the program is feasible. Then, generate programs (F6)."

    return STATUS_OK, message


#if __name__== "__main__":
#    debugging:
#    Part_2_RoboDK(None, "C:/Users/Albert/AppData/Local/Temp/Untitled v1_PART.stl")